__title__ = 'pjrpc'
__description__ = 'Extensible JSON-RPC library'
__url__ = 'https://github.com/dapper91/pjrpc'

__version__ = '1.13.0'

__author__ = 'Dmitry Pershin'
__email__ = 'dapper1291@gmail.com'

__license__ = 'Public Domain License'
